/* -*- mode: c++; c-basic-offset: 3; -*- */
#include "aperture.hh"
#include "lcl_array.hh"
#include "DVecType.hh"
#include "jsonStack.hh"
#include "GDSPlot.hh"
#include "html/document.hh"
#include "html/image.hh"
#include "html/Table2C.hh"
#include "html/writer.hh"
#include "TCanvas.h"
#include <cmath>
#include <cstring>
#include <algorithm>
#include <iostream>
#include <sstream>

#include "poissonge.icc"

using namespace std;
typedef triggers::trig_select trig_select;
typedef triggers::trig_veto   trig_veto;

EXECDAT(aperture)

//======================================  constructor
aperture::aperture(int argc, const char* argv[]) 
: DatEnv(argc, argv), _logCL(0), _logCL1D(0), _corr_limit(0.5), _M(16),
   _osc(getDacc()), _plot_metric(kTriggersOverExpected)
{
   _config = argv[argc-1];
   configure(_config);
   _trig.set_debug(Debug());
   getDacc().setStride(16.0);
}

//======================================  Destructor
aperture::~aperture(void) {
   cout << "Starting post-processing" << endl;
   gps_t start = _channel[0].ref_values().getStartTime().getS();
   gps_t stop  = _channel[0].ref_values().getEndTime().getS();
   Interval dt = _channel[0].ref_values().getTStep();
   size_t nBin = double(stop - start)/double(dt);
   cout << "reading triggers from: ";
   for (auto i=0; i<_trig_path.size(); i++) {
      if (i) cout << ", ";
      cout << _trig_path[i];
   }
   cout << "Trigger selection criteria: " << endl;
   for (size_t i=0; i<_trig.select_list().size(); i++) {
      _trig.select_list()[i].critter(cout);
   }
   cout << " gps range: " << start << "-" << stop << endl;
   _trig.read_triggers(_trig_path, start, stop, dt);

   //-----------------------------------  Veto triggers based on channel values
   cout << "Vetoing triggers" << endl;
   size_t Nc = _channel.size();
   for (size_t i=0; i<Nc; i++) {
      _trig.veto_triggers(_channel[i].name(), _channel[i].ref_values());
   }
   _trig.veto_stats();
   
   cout << "Starting channel pair test" << endl;
   calculate();
}

//======================================  Calculate everything.
void
aperture::calculate(void) {
   //-----------------------------------  Get limits perform 1D tests
   size_t N = _channel.size();
   for (size_t i=0; i<N; i++) {
      if (_channel[i].empty()) {
	 cout << _channel[i].name() << " is empty!" << endl;
	 _channel[i].set_reject(1);
         continue;
      }
      _channel[i].set_limits(_M);

      //--------------------------------  Find worst entry
      channels::chan_stat stat(i);
      bool fail = _channel[i].test1D(_trig.series(), _logCL1D, stat,
				     _reject_outliers);
      if (fail) {
         _stat1d_vector.push_back(stat);
      }
   }
   std::sort(_stat1d_vector.begin(), _stat1d_vector.end());

   //-----------------------------------  get the expected number per cell.
   cout << "start 2D tests..." << endl;
   lcl_array<int> kArray(_M*_M);
   lcl_array<int> tArray(_M*_M);
   for (size_t i=1; i<N; i++) {
      if (_channel[i].test_reject(_reject_outliers)) continue;
      DVectI dvi(*_channel[i].ref_rank().refDVect());
      size_t _nEnt = dvi.size();
      double trigAvg = _trig.average();
      for (size_t j=0; j<i; j++) {
	 if (_channel[j].test_reject(_reject_outliers)) continue;
	 //cout << "computing {" << _channel[i].name() << ", "
	 //     << _channel[j].name() << "}" << endl;

	 //-----------------------------  calculate matrices
	 calc_matrices(i, j, tArray.get(), kArray.get());
	 
	 //-----------------------------  flag stats
	 stat_2d stat(i, j);
	 calc_stats(tArray.get(), kArray.get(), stat);

         if (stat.correlation < _corr_limit && stat.min_likelihood < _logCL) {
	    size_t savij = stat.min_index;
            size_t kj = savij % _M;
            size_t ki = savij / _M;
            cout << _channel[i].name() << "[" << ki << "] " 
                 << _channel[j].name() << "[" << kj << "] N="
		 << tArray[savij]
	         << " expect=" << kArray[savij]*trigAvg 
                 << " pois=" << stat.min_likelihood << endl;
	    cout << "correlation = " << stat.correlation << endl;
	    _stat_vector.push_back(stat);
         }
      }
   }

   //------------------------------------  sort results
   std::sort(_stat_vector.begin(), _stat_vector.end());
   lcl_array<bool> used(N);
   for (size_t i=0; i<N; i++) used[i] = false;
   size_t inxok = 0;
   for (size_t i=0; i<_stat_vector.size(); i++) {
      size_t inxa = _stat_vector[i].channel_a;
      size_t inxb = _stat_vector[i].channel_b;
      if (used[inxa] || used[inxb]) continue;
      if (i != inxok)  _stat_vector[inxok] =  _stat_vector[i];
      inxok++;
      used[inxa] = true;
      used[inxb] = true;
   }
   make_page();
}

//======================================  Formatting utility
inline string
CLstring(double logCL) {
   ostringstream msg;
   msg << logCL << " (log " << exp(logCL) << ")";
   return msg.str();
}

inline const char* 
boolString(bool yn) {
   return yn ? "true" : "false";
}

//======================================  Read the configuration file
void
aperture::configure(const std::string& cfile) {
   jsonStack js(cfile);
   double CL = 0, CL1D = 0;

   //-----------------------------------  Numeric parameters
   js.fetch_data("confidence", CL);
   js.fetch_data("log-confidence", _logCL);
   js.fetch_data("confidence-1d", CL1D);
   js.fetch_data("log-confidence-1d", _logCL1D);
   js.fetch_data("correlation-limit", _corr_limit);
   js.fetch_data("reject-outliers",   _reject_outliers);

   //-----------------------------------  OSC info
   js.fetch_data("osc-file", _oscfile);
   js.fetch_data("osc-cond", _osc_cond);

   //-----------------------------------  html directory
   js.fetch_data("html-directory", _html_directory);   

   //-----------------------------------  Trigger selections.
   js.fetch_data("trigger-path", _trig_path);

   string table_name = "sngl_burst";
   js.fetch_data("trigger-table-name", table_name);
   _trig.set_table_name(table_name);

   string time_field = "peak_time";
   js.fetch_data("trigger-time-field", time_field);
   _trig.set_time_field(time_field);

   string time_ns_field = "peak_time_ns";
   js.fetch_data("trigger-time-ns-field", time_ns_field);
   _trig.set_time_ns_field(time_ns_field);

   bool no_retrigger = true;
   js.fetch_data("inhibit-retrigger", no_retrigger);
   _trig. set_no_retrigger(no_retrigger);
   
   if (js.push_member("trigger-select")) {
      jsonStack::autopop pop_tsel(js);
      while (js.iterate()) {
	 jsonStack::autopop pop_tsiter(js);
         static const char* ts_fields[] = {"field", "string-eq", 
					   "num-ge", "num-lt", 0};
	 string inv_fields = js.validate(ts_fields);
	 if (!inv_fields.empty()) {
	    cerr << "Invalid trigger-select field(s); " << inv_fields << endl;
	    continue;
	 }
         trig_select ts;
         js.fetch_data("field", ts.field);
         if (js.fetch_data("string-eq", ts.string)) {
	    ts.type = trig_select::s_string;
         } else if (js.fetch_data("num-ge", ts.number)) {
	    ts.type = trig_select::s_num_ge;
         } else if (js.fetch_data("num-lt", ts.number)) {
	    ts.type = trig_select::s_num_lt;
         } else {
	    ts.type = trig_select::s_pass;
         }
         _trig.add_select(ts);
      }
   }

   if (js.push_member("vetos")) {
      while (js.iterate()) {
         trig_veto tv;
	 tv.val_min = -1.0/0.0;
	 tv.val_max =  1.0/0.0;
         js.fetch_data("channel", tv.channel);
         js.fetch_data("val-min", tv.val_min);
         js.fetch_data("val-max", tv.val_max);
	 _trig.add_veto(tv);
         js.pop();
      }
      js.pop();
   }

   //-----------------------------------  x-y plot metric
   string plotMetric;
   if (js.fetch_data("plot-metric", plotMetric)) {
      try {
	 _plot_metric = metric_value(plotMetric);
      } catch (runtime_error& e) {
	 cerr << "Unrecognised plot metric: " << plotMetric << endl;
      }
   }

   //-----------------------------------  Channel list
   std::vector<std::string> chan_vect;
   size_t nChan = js.fetch_data("channels", chan_vect);
   if (!nChan) throw runtime_error("aperture: Error reading channel list");
   for (size_t i=0; i<nChan; i++) {
      if (chan_vect[i].empty()) continue;
      _channel.push_back(channels(chan_vect[i]));
      getDacc().addChannel(chan_vect[i]);
   }

   //------------------------------------  Set stuff up
   if (!_oscfile.empty()) _osc.readConfig(_oscfile);
   if (_logCL == 0) {
      if (CL <= 0) CL = 0.002;
      _logCL = log(CL);
   }
   if (_logCL1D == 0) {
      if (CL1D <= 0) CL1D = 0.0002;
      _logCL1D = log(CL1D);
   }
   
   //-----------------------------------  Print
   print_config(cout);
}

//=======================================  Print configuration parameters
void
aperture::print_config(ostream& out) const {
   out << "confidence:            "   << CLstring(_logCL) << endl;
   out << "confidence-1d:         "   << CLstring(_logCL1D) << endl;
   out << "reject-outliers:       "   << boolString(_reject_outliers) << endl;
   out << "plot-metric:           "   << metric_name(_plot_metric) << "\""
       << endl;
   out << "html-directory:        \"" << _html_directory << "\"" << endl;
   out << "osc-file:              \"" << _oscfile << "\"" << endl;
   out << "osc-cond:              \"" << _osc_cond << "\"" << endl;
   out << "trigger-path:          \"" << _trig_path[0] << "\"" << endl;
   for (size_t i=1; i<_trig_path.size(); i++) {
      out << "                       \"" << _trig_path[i] << "\"" << endl;
   }
   out << "trigger-table-name:    \"" << _trig.table_name() << "\"" << endl;
   out << "trigger-time-field:    \"" << _trig.time_field() << "\"" << endl;
   out << "trigger-time-ns-field: \"" << _trig.time_ns_field() << "\"" << endl;
   out << "inhibit-retrigger:     \"" << boolString(_trig.no_retrigger()) 
				       << "\"" << endl;
   out << "trigger selections:    "   << _trig.select_list().size() << endl;
   out << "trigger vetoes:        "   << _trig.veto_list().size() << endl;
   out << "number of channels:    "   << _channel.size() << endl;
}

//======================================  Calculate matrices
void 
aperture::calc_matrices(size_t i, size_t j, int* trig, int* counts) const {
   //-----------------------------------  Zero statistics arrays
   memset(counts, 0, _M*_M*sizeof(int));
   memset(trig,   0, _M*_M*sizeof(int));

   //-----------------------------------  fill matrices
   const DVectI dvi(*_channel[i].ref_rank().refDVect());
   const DVectI dvj(*_channel[j].ref_rank().refDVect());
   const DVectI& dvt = dynamic_cast<const DVectI&>(*_trig.series().refDVect());
   size_t nEnt = dvi.size();
   for (size_t k=0; k<nEnt; k++) {
      size_t ivsj = dvi[k] * _M + dvj[k];
      counts[ivsj]++;
      trig[ivsj] += dvt[k];
   }
}
   
//======================================  Calculate statistics
void 
aperture::calc_stats(const int* trig, const int* counts, stat_2d& s) const {
   double trigAvg = _trig.average();
   int savij = -1;
   double wx = 0;
   double ssq=0;
   for (size_t inxij=0; inxij < _M * _M; inxij++) {
      ssq += pow(counts[inxij], 2);
      double xpect = counts[inxij] * trigAvg;
      double pois  = log(poissonge(xpect, trig[inxij]));
      if (pois < wx) {
	 wx = pois;
	 savij = inxij;
      }
   }
   size_t nEnt = _trig.series().getNSample();
   s.min_likelihood = wx;
   if (savij >= 0) {
      s.correlation = sqrt(_M * ssq / (nEnt *nEnt));
      s.min_index   = savij;
      s.trigs_found = trig[savij];
      s.trigs_expected = counts[savij] * trigAvg;
   }
}
   
//======================================  Make the html output
void
aperture::make_page(void) const {
   if (_html_directory.empty()) return;
   double trigAvg = _trig.average();
   
   //-----------------------------------  set up bin edges.
   lcl_array<double> xybins(_M+1);
   for (size_t i=0; i <= _M; i++) {
      xybins[i] = i;
   }

   //------------------------------------  Start html document
   html::document ape_page("Ape Page");

   //------------------------------------  Build run parameter table
   html::Table2C* argtb = new html::Table2C("Global Parameters", 
					    "Parameter", "Value");
   ape_page.addObject(argtb);
   argtb->addRow("confidence             ", CLstring(_logCL));
   argtb->addRow("confidence-1d          ", CLstring(_logCL1D));
   argtb->addRow("reject-outliers        ", boolString(_reject_outliers));
   argtb->addRow("plot-metric            ", metric_name(_plot_metric));
   argtb->addRow("html-directory         ", _html_directory);
   argtb->addRow("osc-file               ", _oscfile);
   argtb->addRow("osc-cond               ", _osc_cond);
   argtb->addRow("trigger-path           ", _trig_path[0]);
   for (size_t i=1; i<_trig_path.size(); i++) {
      argtb->addRow("                       ", _trig_path[i]);
   }
   argtb->addRow("trigger-table-name     ", _trig.table_name());
   argtb->addRow("trigger-time-field     ", _trig.time_field());
   argtb->addRow("trigger-time-ns-field  ", _trig.time_ns_field());
   argtb->addRow("inhibit-retrigger      ", boolString(_trig.no_retrigger()));
   argtb->addRow("trigger selections     ", _trig.select_string());
   argtb->addRow("trigger vetoes         ", _trig.veto_list().size());
   argtb->addRow("number of channels     ", _channel.size());
   argtb->addRow("start time             ", _trig.series().getStartTime().getS());
   argtb->addRow("duration               ", _trig.series().getInterval());
   argtb->addRow("outlier channels       ", _stat1d_vector.size());
   argtb->addRow("number triggers        ", _trig.total());

   //-------------------------------------  Build 1d results table
   html::table* tab1d = new html::table("Single Channel Outliers");
   ape_page.addObject(tab1d);
   tab1d->addColumn("Channel");
   tab1d->addColumn("Bin");
   tab1d->addColumn("# found/expected");
   tab1d->refHeader(2).setAlign("center");
   tab1d->addColumn("log(Likelihood)");
   tab1d->refHeader(3).setAlign("right");
   for (size_t i=0; i < _stat1d_vector.size(); i++) {
      const channels::chan_stat& s(_stat1d_vector[i]);
      int id = s.channel_id;
      size_t row = tab1d->addRow();
      tab1d->insertData(row, 0, _channel[id].name());
      tab1d->insertData(row, 1, _channel[id].bin_id(s.min_index));
      ostringstream n_navg;
      n_navg << s.trigs_found << "/" << s.trigs_expected;
      tab1d->insertData(row, 2, n_navg.str());
      //tab1d->refCell(row, 2).setAlign("center");
      tab1d->insertData(row, 3, s.min_likelihood);
   }

   //------------------------------------- Build 2d results table
   html::table* tab = new html::table("Statistics");
   ape_page.addObject(tab);
   tab->addColumn("statistics");
   tab->addColumn("plots");

   //-----------------------------------  Loop over plots.
   TCanvas* canvas = new TCanvas("GDSPlot", "Ape Plots");
   GDSPlot gplot(canvas);
   gplot.set_palette("viridis");
   lcl_array<int> counts(_M*_M);
   lcl_array<int> trig(_M*_M);
   size_t N = _stat_vector.size();
   if (N > 50) N = 50;
   for (size_t i=0; i<N; i++) {
      const stat_2d& stat = _stat_vector[i];
      size_t chan_a = stat.channel_a;
      size_t chan_b = stat.channel_b;
      html::Table2C partbl("", "", "");
      partbl.addRow("channel name A    ", _channel[chan_a].name());
      partbl.addRow("channel A bin     ", 
                    _channel[chan_a].bin_id(stat.min_index/_M));
      partbl.addRow("channel name B    ", _channel[chan_b].name());
      partbl.addRow("channel B bin     ", 
                    _channel[chan_b].bin_id(stat.min_index%_M));
      partbl.addRow("correlation       ", stat.correlation);
      partbl.addRow("Triggers found    ", stat.trigs_found);
      partbl.addRow("Triggers expected ", stat.trigs_expected);
      partbl.addRow("log-likelihood    ", stat.min_likelihood);
      tab->addRow();
      tab->insertData(i, 0, partbl);
      gplot.new_plot();
      calc_matrices(chan_a, chan_b, trig, counts);
      double maxz = 0.0;
      lcl_array<double> data(_M*_M);
      for (size_t j=0; j<_M*_M; j++) {
	 double xpect = counts[j] * trigAvg;
	 if (xpect <= 0) {
	    data[j] = 0;
	 } else {
	    switch (_plot_metric) {
	    case kTriggersOverExpected:
	       data[j] = double(trig[j]) / xpect;
	       break;
	    case kTriggers:
	       data[j] = double(trig[j]);
	       break;
	    case kLikelihood:
	       data[j] = -log(poissonge(xpect, trig[j]));
	       break;
	    }
	 }
	 if (data[j] > maxz) maxz = data[j];
      }
      gplot.set_zrange(0.0, maxz);
      gplot.ylabel(_channel[chan_a].name());
      gplot.xlabel(_channel[chan_b].name());
      if (_plot_metric == kLikelihood) {
	 gplot.zlabel("-log(Likelihood)");
      } else {
	 gplot.zlabel(metric_name(_plot_metric));
      }      
      gplot.surf(_M, xybins, _M, xybins, data);
      ostringstream plotname;
      plotname << "aperture_" << chan_a << "_" << chan_b << ".png";
      gplot.print(_html_directory+"/"+plotname.str());
      html::image img(plotname.str());
      tab->insertData(i, 1, img);
   }
   ofstream out((_html_directory + "/index.html").c_str());
   html::writer wout(out);
   ape_page.write(wout);
}

//======================================  Collect data from specified channels
void
aperture::ProcessData(void) {
   cout << "read data data at: " << getDacc().getFillTime() << endl;
   if (!_osc_cond.empty() && !_osc.satisfied(_osc_cond)) return;
   size_t N = _channel.size();
   for (size_t i=0; i<N; i++) {
      const TSeries* ts = getDacc().refData(_channel[i].name());
      if (!ts) {
         cerr << "aperture: no data for channel: " << _channel[i].name() <<endl;
         continue;
      }
      _channel[i].append(*ts);
   }
}

std::string
aperture::metric_name(stats_metric metric) const {
   switch (metric) {
   case kTriggersOverExpected:
      return "Triggers/Expected";
   case kLikelihood:
      return "kLikelihood";
   case kTriggers:
      return "Triggers";
   }
   return "Unknown-metric";
}

aperture::stats_metric
aperture::metric_value(const std::string& metname) const {
   if (metname == metric_name(kTriggersOverExpected)) return kTriggersOverExpected;
   if (metname == metric_name(kLikelihood))           return kLikelihood;
   if (metname == metric_name(kTriggers))             return kTriggers;
   throw runtime_error("unknown metric name");
}
